Require Import Vtac Program List Arith Even Div2 Relations String.
Require Import basis countable lang subs model basiclemmas.
Require Import Classical FunctionalExtensionality.

Open Scope string_scope.

Set Implicit Arguments.


Section morph. (****************************************************************************)

Variables w1 w2 : world.

Definition morph := wst w1 -> wst w2.

Inductive is_morph (m : morph) : Prop :=
  is_morph_intro
    (MNAM : forall a, wnam w1 a -> wnam w2 a)
    (MEXT : forall s s' (EXT1: wext w1 s' s), wext w2 (m s') (m s))
    (MEXTPUB : forall s s' (EP : wextpub w1 s' s), wextpub w2 (m s') (m s))
    (MVAL : forall s r, ltfun (wnam w1) (wval w1 () s r) (wval w2 () (m s) r))
    (MH : forall s r h1 h2 (H12 : wheap w1 () s r h1 h2), wheap w2 () (m s) r h1 h2)
  .

Definition wmext (m' m : morph) : Prop := forall s, wext w2 (m' s) (m s).
Definition wmextpub (m' m : morph) : Prop := forall s, wextpub w2 (m' s) (m s).

End morph. (********************************************************************************)

Hint Constructors is_morph.

Section iso. (******************************************************************************)

Variables w1 w2 : world.

Definition morph_id w : morph w w := fun s => s.

Definition morph_comp w1 w2 w3 (m2 : morph w2 w3) (m1 : morph w1 w2) : morph w1 w3 := fun s => m2 (m1 s).

Lemma morph_id_is_morph : forall w, is_morph (morph_id w).
Proof. by repeat split; eauto 10. Qed.

Structure iso w1 w2 := mkiso
  { iso1 : morph w1 w2
  ; iso2 : morph w2 w1 }.

Inductive is_iso w1 w2 (i : iso w1 w2) : Prop :=
  is_iso_intro
    (M1 : is_morph (iso1 i))
    (M2 : is_morph (iso2 i))
    (COMP12 : forall s, morph_comp (iso1 i) (iso2 i) s = s)
    (COMP21 : forall s, morph_comp (iso2 i) (iso1 i) s = s)
  .

End iso. (**********************************************************************************)

Hint Constructors is_iso.
Hint Unfold morph_id.

Section category_laws.

Lemma morph_composable: forall w1 w2 w3 (m1 : morph w1 w2) (m2 : morph w2 w3)
    (WK1: is_morph m1)
    (WK2: is_morph m2),
  is_morph (morph_comp m2 m1).
Proof.
  intros; destruct (id WK1), (id WK2); unfold morph_comp; econstructor; intros; eauto 10. 
  specialize (MVAL s r); specialize (MVAL0 (m1 s) r); rrdes MVAL; rrdes MVAL0; red; eauto 20.
Qed.

Lemma morph_identity1: forall w1 w2 (m: morph w1 w2),
  morph_comp (morph_id _) m = m.
Proof. intros; extensionality s; reflexivity. Qed.

Lemma morph_identity2: forall w1 w2 (m: morph w1 w2),
  morph_comp m (morph_id _) = m.
Proof. intros; extensionality s; reflexivity. Qed.

End category_laws.

Section equiv. (******************************************************************************)

Variables w1 w2 : world.

Lemma wiso_glob :
  forall w1 w2 (i: iso w1 w2) (LI: is_iso i) g (gOK : wglob w2 g),
    wglob w1 (fun s => g (iso1 i s)).
Proof.
  intros; destruct (id LI); destruct (id M1), (id M2).
  rdes gOK; rhdes GSUP; rhdes MVAL; rhdes MVAL0.
  by repeat (split; eauto).
Qed.

Lemma wiso_exp_int : forall (i : iso w1 w2) (LI : is_iso i) g (gOK: wglob w2 g) s0 s,
  wexp_core w1 (fun s => g (iso1 i s)) (fun s => g (iso1 i s)) (iso2 i s0) (iso2 i s) <9=
  wexp_core w2 g g s0 s.
Proof.
  cofix X; intros; destruct (id LI); destruct (id M1), (id M2).
  rewrite <- (COMP12 s); simpl.
  destruct PR.
  + by vauto.
  + econstructor 2; try eassumption.
    - eby eapply MEXT.
    - by rewrite <- (COMP12 s0); simpl; apply MEXTPUB.
    - by apply MH.
  econstructor 3; try eassumption.
  - eby eapply MEXT.
  - by apply MH.
  - do 2 red; intros; eapply X; clear X; try eapply KREL; try eassumption; eauto.
    + by rewrite <- (COMP21 s'); simpl; apply MEXTPUB0.
    + by apply wiso_glob.
    + by setoid_rewrite (COMP12 s'').
    + by setoid_rewrite (COMP12 s''); eauto.
Qed.

Lemma wexp_iso : forall (i : iso w1 w2) (LI : is_iso i) g (gOK: wglob w2 g) s0 s,
  wexp w1 (fun s => g (iso1 i s)) (iso2 i s0) (iso2 i s) <3=
  wexp w2 g s0 s.
Proof.
  repeat red; intros; eapply wiso_exp_int, PR; eauto using wiso_glob.
  destruct LI; destruct M1, M2; setoid_rewrite (COMP12 s); eauto.
Qed.

Lemma winhabited_iso : forall (i : iso w1 w2) (LI : is_iso i) (INH: winhabited w1), 
  winhabited w2.
Proof.
  repeat intro; destruct (id LI); destruct (id M1).
  destruct (INH _ (wiso_glob LI gOK)); eauto.
Qed.

Lemma wconsistent_iso : forall (i : iso w1 w2) (LI : is_iso i) (CS: wconsistent w1), 
  wconsistent w2.
Proof.
  red; intros; destruct (id LI); destruct M1, M2.
  exploit (CS _ (wiso_glob LI gOK) (iso2 i s) ty e1 e2). 
  - setoid_rewrite (COMP12 s). 
    by rhdes MVAL0; inv APP; econstructor (by eauto). 
  intros; des; eauto 10 using wexp_iso.
Qed.

Theorem weq_iso : forall D G ty e1 e2
   (i : iso w1 w2)
   (LI : is_iso i)
   (EQ1 : weq D G w1 ty e1 e2),
   weq D G w2 ty e1 e2.
Proof.
  intros; rdes EQ1; destructs LI M1 M2; split; [|split]; try red; intros; eauto.
  - eby eapply winhabited_iso.
  - eby eapply wconsistent_iso.
  destruct EQ; constructor; eauto.
  intros; assert (gOK0:=gOK); rdes gOK0.
  eapply wexp_iso; eauto.
  eapply WEXP; eauto using wiso_glob.
  by setoid_rewrite (COMP12 s); eauto.
Qed.

End equiv. (******************************************************************************)

Section sym. (***************************************************************************)

Lemma is_iso_sym : forall (w1 w2 : world) (i : iso w1 w2) (LI : is_iso i),
  is_iso (mkiso (iso2 i) (iso1 i)).
Proof.
  intros; destruct LI; destruct M1, M2.
  by apply is_iso_intro; eauto; apply is_morph_intro; eauto. 
Qed.

End sym. (*******************************************************************************)

Section swap. (***************************************************************************)

Variables w1 w2 : world_local.

Definition swap_iso  :
  iso (wlift (wprod w1 w2)) (wlift (wprod w2 w1)) :=
  {| iso1 := fun s : wst (wlift (wprod w1 w2)) =>
               (fst s, swap (snd s)) : wst (wlift (wprod w2 w1))
   ; iso2 := fun s => (fst s, swap (snd s))
  |}.

Lemma swap_iso_ok: is_iso swap_iso.
Proof.
  intros; econstructor.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro; des; subst; eauto 25.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro; des; subst; eauto 25.
  - intros; destruct s as (? & (? & ?)); reflexivity.
  - intros; destruct s as (? & (? & ?)); reflexivity.
Qed.

End swap. (******************************************************************************)


Section id. (****************************************************************************)

Variables w : world_local.

Definition id_intro_iso : iso (wlift (wprod w wid)) (wlift w) :=
  {| iso1 := fun s : wst (wlift (wprod w wid)) => (fst s, fst(snd s)) : wst (wlift w)
   ; iso2 := fun s => (fst s, (snd s,()))
  |}.

Ltac elim_hemp :=
  repeat match goal with 
  | [H: hplus ?h hemp ?h' |- _] => apply hplus_hemp_rE in H; subst h' 
  | [H: hplus hemp ?h ?h' |- _] => apply hplus_hemp_lE in H; subst h'
  end. 

Lemma id_intro_ok: is_iso id_intro_iso.
Proof.
  intros; econstructor.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 25; []; destruct ty; contradiction.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 25; try contradiction; intros; destruct s; reflexivity.
  - intros; destruct s; reflexivity.
  - intros; destruct s as (? & (? & [])); reflexivity. 
Qed.

Definition id_elim_iso : iso wref (wlift wid) :=
  {| iso1 := fun s : wst wref => (s,()) : wst (wlift wid)
   ; iso2 := fun s => fst s
  |}.

Lemma id_elim_ok: is_iso id_elim_iso.
Proof.
  intros; econstructor.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 20; []; destruct ty; contradiction.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 20; []; destruct ty; contradiction.
  - intros; destruct s as (?&[]); reflexivity.
  - intros; reflexivity.
Qed.

Lemma wglob_wid: forall g (gOK : wglob (wlift w) g) s0,
  wglob (wlift wid) (fun s => g (fst s,s0)).
Proof.
  intros; eapply (wiso_glob (id_intro_ok)) in gOK.
  by eapply wglob_prod2 in gOK; simpls; eauto.
Qed.

End id. (********************************************************************************)

